# Copyright (c) 2010-2011 Joshua Harlan Lifton.
# See LICENSE.txt for details.

"""Plover: Open Source Stenography Software"""

__version__ = '4.0.0.dev8'
__copyright__ = '(C) Open Steno Project'
__url__ = 'http://www.openstenoproject.org/'
__download_url__ = 'http://www.openstenoproject.org/plover'
__credits__ = ['Founded by stenographer Mirabai Knight.\n\nDevelopers:'
               '\nJoshua Lifton',
               'Hesky Fisher',
               'Ted Morin',
               'Benoit Pierre',
               '\nand many more on GitHub: \n<https://github.com/openstenoproject/plover>'
               ]
__license__ = 'GNU General Public License v2 or later (GPLv2+)'
__description__ = 'Open Source Stenography Software'
__long_description__ = """\
Plover is a free open source program intended to bring realtime
stenographic technology not just to stenographers, but also to
hackers, hobbyists, accessibility mavens, and all-around speed demons."""
